/* -----------------------------------------------------------------------------
*
*  npch.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named COPYING.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npch.h"

#include "../npio.h"
#include "../npctrl.h"
#include "npfile.h"
#include "../os/npos.h"
#include <time.h>
//JJ


int getTrackOffset(char** trackScanner, char* propertyName) {
	int offset = 0;
	while (*trackScanner) {
		if (strstr(*trackScanner, propertyName) != NULL) {
			return offset;
		}
		offset++;
		trackScanner++;
	}
	return -1;
}


#define NUMBER_FLOAT_PROPERTIES 33					//9+9+4+4+7 = 33		//zz-JJ
char* floatProperties[NUMBER_FLOAT_PROPERTIES];
void cdsInit() {
	int fpIdx = 0;
	
	floatProperties[fpIdx++] = "translate_x";
	floatProperties[fpIdx++] = "translate_y";
	floatProperties[fpIdx++] = "translate_z";
	floatProperties[fpIdx++] = "rotate_x";
	floatProperties[fpIdx++] = "rotate_y";
	floatProperties[fpIdx++] = "rotate_z";
	floatProperties[fpIdx++] = "scale_x";
	floatProperties[fpIdx++] = "scale_y";
	floatProperties[fpIdx++] = "scale_z";			//9

	floatProperties[fpIdx++] = "translateRate_x";
	floatProperties[fpIdx++] = "translateRate_y";
	floatProperties[fpIdx++] = "translateRate_z";
	floatProperties[fpIdx++] = "rotateRate_x";
	floatProperties[fpIdx++] = "rotateRate_y";
	floatProperties[fpIdx++] = "rotateRate_z";
	floatProperties[fpIdx++] = "scaleRate_x";
	floatProperties[fpIdx++] = "scaleRate_y";
	floatProperties[fpIdx++] = "scaleRate_z";		//9

	floatProperties[fpIdx++] = "tagMode";
	floatProperties[fpIdx++] = "tagOffset_x";
	floatProperties[fpIdx++] = "tagOffset_y";
	floatProperties[fpIdx++] = "tagOffset_z";		//4

	floatProperties[fpIdx++] = "geometry";
	floatProperties[fpIdx++] = "ratio";
	floatProperties[fpIdx++] = "topo";
	floatProperties[fpIdx++] = "facet";				//4

	floatProperties[fpIdx++] = "colorIndex";
	floatProperties[fpIdx++] = "color_r";
	floatProperties[fpIdx++] = "color_g";
	floatProperties[fpIdx++] = "color_b";
	floatProperties[fpIdx++] = "color_a";
	floatProperties[fpIdx++] = "textureID";
	floatProperties[fpIdx++] = "hide";				//7			//zz-JJ-End
}


void mapTrackToAttribute(pData data, int trackOffset, float* attributeValuePtr)
{
	*data->io.ch.nodePropertyMemoryNextLocation++ = attributeValuePtr;
	data->io.ch.propertyTracks[data->io.ch.propertyTrackIndex++] = trackOffset;
	printf("Mapped trackOffset %d\n", trackOffset);
}

void mapTrackOffset(pData data, int trackOffset, NPnodePtr node, char* propertyName)		//zz-JJ
{
	printf("got node %d, now get address of property %s\n", node->id, propertyName);

	if (strcmp(propertyName, "translate_x") == 0) { mapTrackToAttribute(data, trackOffset, &node->translate.x); }
	if (strcmp(propertyName, "translate_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->translate.y); }
	if (strcmp(propertyName, "translate_z") == 0) { mapTrackToAttribute(data, trackOffset, &node->translate.z); }
	if (strcmp(propertyName, "rotate_x") == 0) { mapTrackToAttribute(data, trackOffset, &node->rotate.x); }
	if (strcmp(propertyName, "rotate_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->rotate.y); }
	if (strcmp(propertyName, "rotate_z") == 0) { mapTrackToAttribute(data, trackOffset,  &node->rotate.z); }
	if (strcmp(propertyName, "scale_x") == 0) { mapTrackToAttribute(data, trackOffset,  &node->scale.x); }
	if (strcmp(propertyName, "scale_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->scale.y); }
	if (strcmp(propertyName, "scale_z") == 0) { mapTrackToAttribute(data, trackOffset, &node->scale.z); }

	if (strcmp(propertyName, "translate_rate_x") == 0) { mapTrackToAttribute(data, trackOffset, &node->translateRate.x); }
	if (strcmp(propertyName, "translate_rate_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->translateRate.y); }
	if (strcmp(propertyName, "translate_rate_z") == 0) { mapTrackToAttribute(data, trackOffset, &node->translateRate.z); }
	if (strcmp(propertyName, "rotate_rate_x") == 0) { mapTrackToAttribute(data, trackOffset, &node->rotateRate.x); }
	if (strcmp(propertyName, "rotate_rate_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->rotateRate.y); }
	if (strcmp(propertyName, "rotate_rate_z") == 0) { mapTrackToAttribute(data, trackOffset, &node->rotateRate.z); }
	if (strcmp(propertyName, "scale_rate_x") == 0) { mapTrackToAttribute(data, trackOffset,  &node->scaleRate.x); }
	if (strcmp(propertyName, "scale_rate_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->scaleRate.y); }
	if (strcmp(propertyName, "scale_rate_z") == 0) { mapTrackToAttribute(data, trackOffset, &node->scaleRate.z); }

//	if (strcmp(propertyName, "tag_mode") == 0) { mapTrackToAttribute(data, trackOffset, &node->tagMode); }
	if (strcmp(propertyName, "tag_offset_x") == 0) { mapTrackToAttribute(data, trackOffset, &node->tagOffset.x); }
	if (strcmp(propertyName, "tag_offset_y") == 0) { mapTrackToAttribute(data, trackOffset, &node->tagOffset.y); }
	if (strcmp(propertyName, "tag_offset_z") == 0) { mapTrackToAttribute(data, trackOffset,  &node->tagOffset.z); }

//	if (strcmp(propertyName, "geometry") == 0) { mapTrackToAttribute(data, trackOffset,  &node->geometry); }
	if (strcmp(propertyName, "ratio") == 0) { mapTrackToAttribute(data, trackOffset,  &node->ratio); }
/*	if (strcmp(propertyName, "topo") == 0) { mapTrackToAttribute(data, trackOffset,  &node->topo); }
	if (strcmp(propertyName, "facet") == 0) { mapTrackToAttribute(data, trackOffset,  &node->facet); }

	if (strcmp(propertyName, "colorIndex") == 0) { mapTrackToAttribute(data, trackOffset,  &node->colorIndex); }
	if (strcmp(propertyName, "color_r") == 0) { mapTrackToAttribute(data, trackOffset,  &node->color.r); }
	if (strcmp(propertyName, "color_g") == 0) { mapTrackToAttribute(data, trackOffset,  &node->color.g); }
	if (strcmp(propertyName, "color_b") == 0) { mapTrackToAttribute(data, trackOffset,  &node->color.b); }
	if (strcmp(propertyName, "color_a") == 0) { mapTrackToAttribute(data, trackOffset,  &node->color.a); }
*/
}		//zz-JJ-End

float** createTracks(void *dataRef, int numberTracks) {
	pData data = (pData) dataRef;
	int i = 0;
	float** trackDataScanner;
	char* trackBuffer;
	trackDataScanner = data->io.ch.trackData = (float**)malloc(sizeof(float*)*(numberTracks+1));

	for (i = 0; i < numberTracks; i++) {
		trackBuffer = (char*)malloc(sizeof(float*)*kNPtrackBufferSize);
		memset(trackBuffer, 0, sizeof(float*)*kNPtrackBufferSize);
		*(trackDataScanner++) = (float*)trackBuffer;
		*(trackDataScanner) = NULL;
	}
	return data->io.ch.trackData;
}

void saveHeader(int offset, void *dataRef, char* header) {
	pData data = (pData) dataRef;
	char* headerCopy = (char*)malloc(strlen(header)+1);
	strcpy(headerCopy, header);
	*(data->io.ch.trackNames + offset) = headerCopy;
	offset++;
	*(data->io.ch.trackNames + offset) = NULL;
	printf("Just saved header %s, now have %d headers\n", header, offset);
	// If we've reached limit, need to re-allocate both lists
}

int saveHeaders(void *dataRef, char** headers) {
	// NULL terminated list of pointers
	int offset = 0;
	while (*headers != NULL) {
		saveHeader(offset++, dataRef, *headers++);
	}
	return offset;
}

// saves data for each track into the different track data buffers
bool saveData(int numberTracks, int offset, float** trackData, char** data) {
	int i = 0;
	float* singleTrack = NULL;
	if (data == NULL) return false;
	for (i = 0; i < numberTracks; i++) {
		singleTrack = *(trackData + i);
		*(singleTrack + offset) = (float)atof(*data++);
	}
	return true;
}



/////////////////////////////////
// end of support functions
/////////////////////////////////

//init variables, allocate RAM
//------------------------------------------------------------------------------
void npInitCh (void* dataRef)
{
	pData data = (pData) dataRef;
	int i = 0;
	data->io.ch.xrefIndex = 0;
	for (i = 0; i < kNPmaxTrackToAttributeMappings; i++) {
		data->io.ch.xref[i].attributeName = NULL;
		data->io.ch.xref[i].trackOffset = 0;
	}
	data->io.ch.updateData = true;
	data->io.ch.channelReadIndex = kNPtrackBufferSize - 1;
	data->io.ch.channelWriteIndex = 0;
	data->io.ch.channelSampleRate = 50;
	data->io.ch.channelDataSize = 1;
	data->io.ch.channelNextDataChangeTime = clock() + 1000*60*60*24*365;  // wait a year before spontaneously starting
	data->io.ch.numberTracks = 0;
	data->io.ch.trackNames = (char**)malloc((kNPmaxTracks+1) * sizeof(char*));  // +1 for NULL pointer indicating end of list
	data->io.ch.nodePropertyMemoryNextLocation = &data->io.ch.nodePropertyMemoryLocations[0];
	data->io.ch.propertyTrackIndex = 0;
	data->io.ch.trackData = NULL;
	for (i = 0; i < kNPmaxFloatProperties; i++) {
		data->io.ch.propertyTracks[i] = 0;
		data->io.ch.nodePropertyMemoryLocations[i] = NULL;
	}
	data->io.ch.pause = 0;
	cdsInit();
}

//------------------------------------------------------------------------------
void npCloseCh (void* dataRef)		//clean-up
{
	return;
}

void dumpChannel(char* s, NPch* channel) {
	printf("%s\n", s);
	printf(" channelNextDataChangeTime: %d\n", channel->channelNextDataChangeTime);
	printf(" channelReadIndex: %d\n", channel->channelReadIndex);
	printf(" channelSampleRate: %d\n", channel->channelSampleRate);
	printf(" channelWriteIndex: %d\n", channel->channelWriteIndex);
	//printf(" : %d\n", channel->nodePropertyMemoryLocations);
	//printf(" : %d\n", channel->nodePropertyMemoryNextLocation);
	printf(" numberTracks: %d\n", channel->numberTracks);
	printf(" propertyTrackIndex: %d\n", channel->propertyTrackIndex);
	//printf(" : %d\n", channel->propertyTracks);
	//printf(" : %d\n", channel->trackData);
	//printf(" : %d\n", channel->trackNames);
}


//
//  Currently have two options for behavior when we reach the end:
//
//  - pause
//  - loop
//
//  If threads are supported this function will be called in its own thread.  In that case, it will
//  implement circular buffer as part of loading data -- waiting for data to be read before overwriting it.
//  With looping with files larger than the circular buffer size, those large files are re-opened and re-read.
//  This allows looping to occur with any size file.
//
//  If threads are not supported, the buffer loading stops as soon as the buffer is full.  The remaining
//  data is never loaded.  Looping occurs only on the data loaded into the buffer.
//
bool loopAtEndOfTrack = true;	//zz,  move to data->io.ch
void loadTracksIntoCircularBuffer (void* dataRef)
{
	pData data = (pData)dataRef;
	pNPch ch = &data->io.ch;
	FILE* fp = ch->fileDataSource;

	while ( fp != NULL ) {
		while (saveData(data->io.ch.numberTracks, data->io.ch.channelWriteIndex++, data->io.ch.trackData, npFileTokens(fp))) {
			if (data->io.ch.channelWriteIndex > data->io.ch.channelDataSize) data->io.ch.channelDataSize = data->io.ch.channelWriteIndex;

			// circular buffer requires thread, otherwise, we are totally done
			if (nposSupportsAntzThreads()) {
				if (data->io.ch.channelWriteIndex == kNPtrackBufferSize) data->io.ch.channelWriteIndex = 0;
				//printf("channelReadIndex %d, channelWriteIndex %d, channelDataSize now %d\n", data->io.ch.channelReadIndex, data->io.ch.channelWriteIndex, data->io.ch.channelDataSize);
				while (data->io.ch.channelWriteIndex == data->io.ch.channelReadIndex) { // wait for data to be read before overwriting
					printf("Waiting for data to be read..\n");
					_sleep(100);
				}
			}
			// without threads, break out
			else if (data->io.ch.channelWriteIndex == kNPtrackBufferSize) break;
		}
		npFileClose(fp, dataRef);
		fp = data->io.ch.fileDataSource = NULL;

		// if we are looping, and the buffer cycled, re-read the file
		if ((loopAtEndOfTrack) && (data->io.ch.channelDataSize == kNPtrackBufferSize) && nposSupportsAntzThreads()) {
			data->io.ch.channelReadIndex = kNPtrackBufferSize - 1;
			data->io.ch.channelWriteIndex = 0;
			fp = data->io.ch.fileDataSource = npFileOpen(data->io.ch.fullFilePath, "r", dataRef);
		}
	}
	nposEndThread();
}

//open the file and read contents into a buffer
//determine the file type, and call either npChMapImport() or npChDataImport()
//------------------------------------------------------------------------------
//
//  The channel is the data pipe between the tracks and the nodes
//
//  Format is:
//
//     id,track_set_id,track_column,attribute_name,id1,id2,id3
//
//  For now, only using track_column, attribute_name
//
//  track_column is the column offset in the data loaded with npLoadChTracks
//  
//  data loaded here goes into xref[] array of NPxref that map a track offset to an attribute name
//
void npLoadChMap (const char* filePath, void* dataRef)
{
	FILE* fp;
	char** tokens;
	char** tokenScanner;
	int tokenCount = 0;
	pData data = (pData) dataRef;
	char path[2048];
	char* fullFilePath = NULL;
	int size;

	npGetCWD (path, &size);
	fullFilePath = strcat(path, filePath);
	fp = npFileOpen(fullFilePath, "r", dataRef);

	if (fp != NULL) {
		tokens = tokenScanner = npFileTokens(fp);
		// expecting 7 columns, if we don't get that report error
		while (*tokenScanner++) tokenCount++;
		if (tokenCount != 7) { perror("Invalid channel file, unexpected number of tokens"); return; }

		// first line was header, now start with data
		tokens = npFileTokens(fp);
		while (tokens != NULL) {
			data->io.ch.xref[data->io.ch.xrefIndex].attributeName = newstrcpy(*(tokens + kNPattributeNameOffset));
			data->io.ch.xref[data->io.ch.xrefIndex].trackOffset = atoi(*(tokens + kNPtrackOffset));
			printf("Mapped track offset %d to attribute %s\n", data->io.ch.xref[data->io.ch.xrefIndex].trackOffset, data->io.ch.xref[data->io.ch.xrefIndex].attributeName);
			if (data->io.ch.xrefIndex < kNPmaxTrackToAttributeMappings) data->io.ch.xrefIndex++;
			tokens = npFileTokens(fp);
		}
	}
}

//
//  The track file is a text form of multiple tracks
//
//  The first CSV line is column headers, free form text, not used by program, can be used to describe track
//
//  Each column of data is a single track
//
void npLoadChTracks (const char* filePath, void* dataRef)
{
	int size = 0;
	char path[2048];
	char* fullFilePath = NULL;
	FILE* fp = NULL;
	float** trackData;
	int numberTracks = 0;
	pData data = (pData) dataRef;

	npGetCWD (path, &size);
	fullFilePath = newstrcpy(strcat(path, filePath));
	fp = npFileOpen(fullFilePath, "r", dataRef);
	if (fp != NULL) {
		data->io.ch.fullFilePath = fullFilePath;
		data->io.ch.fileDataSource = fp;
		data->io.ch.numberTracks = numberTracks = saveHeaders(dataRef, npFileTokens(fp));
		trackData = createTracks(dataRef, numberTracks);

		// data loaded in another thread to allow circular buffer, new thread is for filling the buffer from the file
		// this thread reads from circular buffer, the new thread we are creating here writes to the circular buffer
		// both reading and writing threads follow boundary convention to implement circular buffer
		// 
		// circular buffer conventions:
		//   writeIndex != readIndex means data can be written at the current index
		//   writeIndex == readIndex means the circular buffer is full, nothing more can be written into it
		//   ((readIndex + 1)%circularBufferSize) == writeIndex means there is no data available to read
		//
		nposBeginThread(loadTracksIntoCircularBuffer, dataRef);
	} else {
		printf("Failed opening %s\n", fullFilePath);
	}
}

//passed in a buffer containing the entire contents of the antzchmap000x.csv
//parses the CSV formatted contents and builds the channel map list
//merging by using multiple imports allowed
//------------------------------------------------------------------------------
void npChMapImport (char* buffer, int size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataImport (char* buffer, int size, void* dataRef)	//antzch0001.csv
{
	return;
}

//exports entire channel map to the buffer in CSV format
//the function that calls this is responsible for freeing the buffer when done
//typically would be written to a CSV file or DB then the buffer is freed
//------------------------------------------------------------------------------
void npChMapExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//all functions can be used to read/write to or from a CSV file or DB
//separate out the buffer to make this possible...
//add or remove ch attribute track pairs
//------------------------------------------------------------------------------
void npChMapAttributeAdd (int channel, int attribute, int track, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChMapAttributeDelete (int channel, int attribute, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npNewChannel (int channel, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npDeleteChannel (int channel, void* dataRef)
{
	return;
}

								//will remove '.x' from node->channel.x debug zz
//the node->channe.x specifies the channel map to use for attribute mapping
//of the track-attrib pairs added to the ptr-ptr list
//chID = 0 is the null channel, results in no channel mapping
//------------------------------------------------------------------------------
void npChSubscribeNode (NPnodePtr node, void* dataRef)
{
	int i = 0;
	pData data = (pData) dataRef;

	if (node->channel.x == 0)	//zz
		return;

	data->io.ch.channelNextDataChangeTime = clock();
	data->io.ch.channelReadIndex = 0;
	for (i = 0; i < data->io.ch.xrefIndex; i++) {
		mapTrackOffset(data, data->io.ch.xref[i].trackOffset, node, data->io.ch.xref[i].attributeName);
	}
	printf("npChSubscribeNode, nodePropertyMemoryScanner starts at 0x%x, data->io.ch.nodePropertyMemoryNextLocation is 0%x\n",
		&data->io.ch.nodePropertyMemoryLocations[0], data->io.ch.nodePropertyMemoryNextLocation);
}

void npUpdateNodes(void *dataRef) {
	pData data = (pData) dataRef;
	float **nodePropertyMemoryScanner =  &data->io.ch.nodePropertyMemoryLocations[0];
	float *track = NULL;
	int channelScanIdx = 0;
	if (data->io.ch.updateData) {
		data->io.ch.updateData = false;
		while (nodePropertyMemoryScanner != data->io.ch.nodePropertyMemoryNextLocation) {
			track = *(data->io.ch.trackData + data->io.ch.propertyTracks[channelScanIdx]);
			**nodePropertyMemoryScanner++ = *(track + data->io.ch.channelReadIndex);
			channelScanIdx++;
		}
	}
}

//removes any pointers in the list associated with the node passed in
//used when changing channels which first removes the node, 
//then changes the chID and calls npChSubscribeNode to setup the new channel ptrs
//also called by npNodeDelete()
//------------------------------------------------------------------------------
void npChRemoveNode (NPnodePtr node, void* dataRef)
{
	pData data = (pData) dataRef;
	char* nodeStart = (char*)node;
	char* nodeEnd = (nodeStart + sizeof(struct NPnode));
	float **nodePropertyMemoryScanner =  &data->io.ch.nodePropertyMemoryLocations[0];
	while (nodePropertyMemoryScanner != data->io.ch.nodePropertyMemoryNextLocation) {
		if ((*nodePropertyMemoryScanner >= (float*)nodeStart) && (*nodePropertyMemoryScanner < (float*)nodeEnd)) {
			data->io.ch.nodePropertyMemoryNextLocation--;
			*nodePropertyMemoryScanner =*(data->io.ch.nodePropertyMemoryNextLocation);
			*(data->io.ch.nodePropertyMemoryNextLocation) = NULL;
		} else {
			nodePropertyMemoryScanner++;
		}
	}
	printf("npChRemoveNode, nodePropertyMemoryScanner starts at 0x%x, data->io.ch.nodePropertyMemoryNextLocation is 0%x\n",
		&data->io.ch.nodePropertyMemoryLocations[0], data->io.ch.nodePropertyMemoryNextLocation);
}


//(un)freeze a node in the list (or should we just add and del..) debug zz

//typically called once per cycle
//1st, update new input stream data, copy the last cycle into the track buffers
//2nd, update/calculate logical tracks, they may use raw tracks as inputs
//3rd, update nodes using track to node-attrib ptr lists
//lastly, output streams, such as live IO devices buffers or to a file
//------------------------------------------------------------------------------
void npUpdateCh (void* dataRef)
{
	pData data = (pData) dataRef;
	clock_t now = clock();
	int nextReadIndex;

	if (data->io.ch.pause) return;  // nothing to update if we have paused
	if (now > data->io.ch.channelNextDataChangeTime) {
		//printf("time for update... channelReadIndex %d, channelWriteIndex %d, channelDataSize %d\n", data->io.ch.channelReadIndex, data->io.ch.channelWriteIndex, data->io.ch.channelDataSize);

		data->io.ch.channelNextDataChangeTime += data->io.ch.channelSampleRate;
		nextReadIndex = (data->io.ch.channelReadIndex + 1) % (data->io.ch.channelDataSize);

		// if there is data to read, update the read index, and set the flag that causes the updates
		if (nextReadIndex != data->io.ch.channelWriteIndex) {
			data->io.ch.updateData = true;
			data->io.ch.channelReadIndex = nextReadIndex;
		printf("Reading %d, channelWriteIndex %d, channelDataSize %d\n", data->io.ch.channelReadIndex, data->io.ch.channelWriteIndex, data->io.ch.channelDataSize);
		}
	}

	npUpdateNodes(data);	//zz-JJ
}


//--- will add some stream functions
//multiple types DB, CSV, Network UDP & IP, LiveIO - audio, video, TNG3B, mouse
//updated with npUpdateCh() 
//------------------------------------------------------------------------------
void npChOpenInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChOpenOutputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseOutStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npchPause (void* dataRef)	//zz-JJ
{
	char msg[256];
	pData data = (pData) dataRef;

	// reset next time to change, needed when resuming
	data->io.ch.channelNextDataChangeTime = clock();
	sprintf(msg, "pause: %d\n", data->io.ch.pause);
	npPostMessage (msg, kNPmessageTypeCtrl, dataRef);

	return;
}

//zz-JJ
//------------------------------------------------------------------------------
void npSyncChMap (void* dataRef)
{
	//delete entire channel pointer map list
	//add all nodes with channels back to the pointer map list
	npTraverseMap (npChSubscribeNode, dataRef);
}

